"""
A basic version of chaise that just uses enriched dictionaries as the documents.

:class:`BasicLoader`, :class:`BasicSession`, and :class:`BasicPool` do not do
any migration handling.

:class:`DictRegistry` should be paired with :class:`~chaise.CouchSession` and
:class:`~chaise.SessionPool` if you want some basic document type handling.
"""

from . import DocumentRegistry, CouchSession, SessionPool


class Document(dict):
    """
    A dictionary with some CouchDB attributes
    """

    #: Document ID
    id: str | None = None

    #: Document revision
    rev: str | None = None

    #: Has the document been deleted? (ie, is this a tombstone?)
    deleted: bool = False

    #: Attachment information, if requested
    attachments: dict | None = None

    #: List of conflicts, if requested
    conflicts: list | None = None

    # List of deleted conflicts, if requested
    deleted_conflicts: list | None = None

    #:
    local_seq: str | None = None

    #:
    revs_info: list | None = None

    #:
    revisions: dict | None = None

    def __repr__(self):
        return f"<{type(self).__name__} {self.id!r}@{self.rev!r} {' '.join(f'{k}={v!r}' for k,v in self.items())}>"


class BasicLoader:
    """
    Provides loading without worrying about types or migrations.
    """

    def load_from_blob(self, blob: dict, _kind=Document) -> Document:
        doc = _kind()
        doc.id = blob.pop("_id", None)
        doc.rev = blob.pop("_rev", None)
        doc.deleted = blob.pop("_deleted", False)
        doc.attachments = blob.pop("_attachments", None)
        doc.conflicts = blob.pop("_conflicts", None)
        doc.deleted_conflicts = blob.pop("_deleted_conflicts", None)
        doc.local_seq = blob.pop("_local_seq", None)
        doc.revs_info = blob.pop("_revs_info", None)
        doc.revisions = blob.pop("_revisions", None)
        doc.update(blob)
        return doc

    def dump_to_blob(self, doc: Document) -> dict:
        return (
            doc
            | ({"_id": doc.id} if doc.id is not None else {})
            | ({"_rev": doc.rev} if doc.rev is not None else {})
            | {
                "_deleted": doc.deleted,
                # The rest of it is informational not editable directly
            }
        )

    def update_doc(self, doc, **fields):
        assert all(k.startswith("_") for k in fields)
        for k, v in fields.items():
            setattr(doc, k.removeprefix("_"), v)


class BasicSession(CouchSession):
    loader = BasicLoader


class BasicPool(SessionPool):
    session_class = BasicSession


class DictRegistry(DocumentRegistry):
    _loader = BasicLoader()

    def load_doc(self, cls: type[Document], blob: dict) -> Document:
        """
        Load a document
        """
        return self._loader.load_from_blob(blob, _kind=cls)

    def dump_doc(self, doc: Document) -> dict:
        """
        Save a document
        """
        return self._loader.dump_to_blob(doc)

    def update_doc(self, doc, **fields):
        self._loader.update_doc(doc, **fields)



"""
Various data structures used by chaise.
"""

import dataclasses
import enum
import functools
import operator

import chaise  # Be careful using this, for circular import reasons


@dataclasses.dataclass
class AllDocs_DocRef:
    """
    A document reference returned by :meth:`~chaise.Database.iter_all_docs`
    """

    #: The document ID
    docid: str

    #: The revision of the document
    rev: str

    _db: "chaise.Database"
    _doc: object | None

    async def doc(self):
        """
        Actually get the document. Since docid+rev is roughly immutable, caches
        it.

        (Might be pre-loaded by the producing function.)
        """
        if self._doc is None:
            # docid + revision is immutable(ish), so it's safe to cache
            # (It can still be deleted/vacuumed, but that's fine. probably.)
            self._doc = await self._db.get(self.docid, rev=self.rev)
        return self._doc


class AscDesc(enum.StrEnum):
    ASC = "asc"
    DESC = "desc"


@dataclasses.dataclass
class IndexDef:
    """
    Index Definition.
    """

    #: Fields and their direction.
    fields: dict[str, AscDesc]


@dataclasses.dataclass
class Index:
    """
    Return of :meth:`~chaise.Database.iter_indexes`
    """

    #: ID of the design document the index belongs to.
    ddoc: str
    #: Name of the index.
    name: str
    #: Partitioned (:const:`True`) or global (:const:`False`) index.
    partitioned: bool
    #: Type of the index. Currently ``"json"`` is the only supported type.
    type: str
    #: Definition of the index, containing the indexed fields and the sort
    #: order: ascending or descending.
    def_: IndexDef

    @classmethod
    def from_dict(cls, data):
        return cls(
            ddoc=data["ddoc"].removeprefix("_design/"),
            name=data["name"],
            partitioned=data["partitioned"],
            type=data["type"],
            def_=IndexDef(
                fields={
                    k: AscDesc(v)
                    for k, v in functools.reduce(
                        operator.or_, data["def"]["fields"], {}
                    ).items()
                },
            ),
        )



"""
Deals with finding and loading database/ddoc description files.
"""

import dataclasses
import importlib.resources
import posixpath
from typing import Iterable, Protocol

import ckdl


class ParseError(Exception):
    pass


@dataclasses.dataclass
class Index:
    name: str
    fields: list[str]
    filter: dict | None = None


@dataclasses.dataclass
class DesignDoc:
    name: str
    _resource: importlib.resources.abc.Traversable
    indexes: list[Index]


@dataclasses.dataclass
class DeclaredDB:
    name: str
    _resource: importlib.resources.abc.Traversable
    # options: ...
    ddocs: list[DesignDoc]


class Loader(Protocol):
    def load_database(self, file_contents: str, db: DeclaredDB): ...

    def load_ddoc(self, file_contents: str, ddoc: DesignDoc): ...


class KdlLoader:
    def load_database(self, file_contents: str, db: DeclaredDB):
        if not file_contents.strip():
            return
        doc = ckdl.parse(file_contents, version="detect")
        assert len(doc.nodes) == 1
        (node,) = doc.nodes
        assert node.name == "database"
        assert not node.children
        match node.args:
            case ():
                pass
            case (name,):
                db.name = str(name)
            case _:
                raise ParseError(f"Invalid KDL: Unexpected arguments: {node}")

    def load_ddoc(self, file_contents: str, ddoc: DesignDoc):
        if not file_contents.strip():
            return
        doc = ckdl.parse(file_contents, version="detect")

        for node in doc.nodes:
            match node.name:
                case "index":
                    ddoc.indexes.append(self._parse_index(node))
                case _:
                    raise ParseError(f"Uknown node: {node}")

    def _parse_index(self, node) -> Index:
        match node.args:
            case ():
                raise ParseError(f"Invalid KDL: Expecting arguments: {node}")
            case (name,):
                index_name = name
            case _:
                raise ParseError(f"Invalid KDL: Unexpected arguments: {node}")

        index_fields = []
        for child in node.children:
            match child.name:
                case "field" | "fields":
                    index_fields += child.args
                case _:
                    raise ParseError(f"Invalid KDL: Uknown node: {child}")

        return Index(
            name=index_name,
            fields=index_fields,
        )


LOADERS: dict[str, type[Loader]] = {
    ".kdl": KdlLoader,
    # ".json": ...,
    # ".yaml": ...,
    # ".yml": ...,
    # ".toml": ...,
    # '.py': ...,
}


def walk(anchor):
    """
    Generates (<path>, traversible)
    """

    def _recurse(base, trav):
        yield base, trav
        if trav.is_dir():
            for t in trav.iterdir():
                if base:
                    n = posixpath.join(base, t.name)
                else:
                    n = t.name
                if not hasattr(t, "parent"):
                    t.parent = trav
                yield from _recurse(n, t)

    yield from _recurse("", importlib.resources.files(anchor))


def find_dbs(anchor) -> Iterable[DeclaredDB]:
    for path, t in walk(anchor):
        # Traversable is a subset of pathlib.Path--it exludes most of the name
        # manipulation features
        parent, base = posixpath.split(path)
        stem, ext = posixpath.splitext(base)
        if stem == "__db__":
            try:
                loader = LOADERS[ext]()
            except KeyError as exc:
                raise Exception(f"Unable to determine loader for {path}") from exc
            else:
                obj = DeclaredDB(
                    name=posixpath.basename(parent),
                    _resource=t,
                    ddocs=[],
                )
                loader.load_database(t.read_text(), obj)
                obj.ddocs += list(find_ddocs(t.parent))
                yield obj


def find_ddocs(trav):
    assert trav.is_dir()
    for file in trav.iterdir():
        if file.is_dir():
            continue
        stem, ext = posixpath.splitext(file.name)
        if stem == "__db__":
            continue
        try:
            loader = LOADERS[ext]()
        except KeyError as exc:
            # TODO: emit warning
            exc
        else:
            ddoc = DesignDoc(
                name=stem,
                _resource=file,
                indexes=[],
            )
            loader.load_ddoc(file.read_text(), ddoc)
            yield ddoc



"""
CouchDB management tool
"""

import argparse
import anyio
import logging
import os

from .. import Missing
from .datafiles import find_dbs
from .client import ConstantPool


async def apply(args):
    """
    Apply the given database descriptions
    """
    session = await ConstantPool(args.server).session()

    # Ensure all the dbs exist
    for wanted in find_dbs(args.module):
        try:
            livedb = await session.get_db(wanted.name)
        except Missing:
            print(f"Creating {wanted.name}...")  # TODO: Logging
            livedb = await session.create_db(wanted.name)
        else:
            print(f"{wanted.name} exists")

        current_indexes = {(i.ddoc, i.name): i async for i in livedb.iter_indexes()}
        wanted_indexes = {
            (ddoc.name, i.name): i for ddoc in wanted.ddocs for i in ddoc.indexes
        }

        to_make = set(wanted_indexes.keys()) - set(current_indexes.keys())
        for dname, iname in to_make:
            idx = wanted_indexes[dname, iname]
            await livedb.add_index(
                ddoc=dname,
                name=iname,
                fields=idx.fields,
            )
        # TODO: Delete indexes
        # TODO: Update indexes


def _arg_parser():
    async def usage(args):
        parser.print_usage()

    parser = argparse.ArgumentParser(
        prog="chaise",
        description=__doc__,
    )
    parser.add_argument(
        "--verbose", action="store_true", help="Enable verbose logging."
    )
    parser.add_argument(
        "--server",
        metavar="URL",
        help="Server URL to use (or $COUCHDB_URL)",
        default=...,
    )
    parser.set_defaults(func=usage)

    subparsers = parser.add_subparsers(title="Subcommands")

    applyp = subparsers.add_parser("apply", help=apply.__doc__)
    applyp.set_defaults(func=apply)
    applyp.add_argument("module")

    return parser


async def main():
    args = _arg_parser().parse_args()

    if args.server is ...:
        args.server = os.environ.get("COUCHDB_URL", None)
    elif not args.server:
        args.server = None

    logging.basicConfig(
        level=logging.INFO if args.verbose else logging.WARNING,
        format="%(name)s %(levelname)s %(message)s",
    )

    await args.func(args)


def entry():
    anyio.run(main)



import json
import typing
from typing import (
    AsyncIterator,
    Literal,
    Callable,
    Protocol,
    TypeVar,
    Generic,
    ClassVar,
    Any,
)
import warnings

import httpx

from . import structs


DOCT = TypeVar("DOCT")

# 3.12: type TypeIDType = ...
# Technically, anything JSONable is allowable, but only allowing atomic types
# makes a bunch of reasoning easier.
#: The type of class identifiers
TypeIDType = str | int | bool | None


class DocumentLoader(Protocol, Generic[DOCT]):
    """
    Protocol for things that go between JSON blobs and hydrated document
    instances.
    """

    def load_from_blob(self, blob: dict) -> DOCT:
        """
        Convert a JSON blob into a document object.
        """

    def dump_to_blob(self, doc: DOCT) -> dict:
        """
        Convert a document into a JSON blob.
        """

    def update_doc(self, doc: DOCT, **fields):
        """
        Update a document object in-place.
        """


class DocumentRegistry:
    """
    Handles de/serialization, manages migrations, etc.

    Default loader implementation.

    Do not use directly. You probably want one of the :ref:`integrations`.
    """

    TYPE_KEY = ""

    _docclasses: ClassVar[dict[TypeIDType, type]] = {}
    _migrations: ClassVar[list[tuple[TypeIDType, TypeIDType, Callable]]] = []

    def __init_sublcass__(cls):
        cls._docclasses = {}
        cls._migrations = []

    @classmethod
    def _get_class_from_name(cls, name: TypeIDType) -> type:
        return cls._docclasses[name]

    @classmethod
    def _get_name_from_class(cls, klass: type) -> TypeIDType:
        for name, kind in cls._docclasses.items():
            if issubclass(klass, kind):  # In case of decorator shenanigans
                return name
        else:
            raise ValueError(f"Couldn't find name for {klass}")

    @classmethod
    def document(cls, name: TypeIDType):
        """
        Register a class as a loadable couch document.

        Args:
            name: The type identifier to save to CouchDB

        .. note::

           The given name must be globally unique and must never change.
        """
        assert not isinstance(name, type)
        assert name not in cls._docclasses

        def _(klass: type):
            cls._docclasses[name] = klass
            return klass

        return _

    @classmethod
    def migration(cls, before: type, after: type):
        """
        Define a function that'll convert between documents.
        """
        # Normalize to the document classes previously registered
        bname = cls._get_name_from_class(before)
        aname = cls._get_name_from_class(after)
        # Enforce linearity
        assert not any(b == bname for b, _, _ in cls._migrations)

        def _(func: Callable):
            cls._migrations.append((bname, aname, func))
            return func

        return _

    def load_doc(self, cls: type, blob: dict):
        """
        Converts a JSON blob into a document.

        Override me.
        """
        raise NotImplementedError

    def dump_doc(self, doc) -> dict:
        """
        Convert a document into a JSON blob.

        Override me.
        """
        raise NotImplementedError

    def update_doc(self, doc, **fields):
        """
        Update a doc in-place
        """
        raise NotImplementedError

    def _migrate(self, bname, doc):
        while funcs := [f for b, _, f in self._migrations if b == bname]:
            (func,) = funcs
            doc = func(doc)
            bname = self._get_name_from_class(type(doc))
        return doc

    def load_from_blob(self, blob):
        type = blob.pop(self.TYPE_KEY)
        klass = self._get_class_from_name(type)
        doc = self.load_doc(klass, blob)
        doc = self._migrate(type, doc)
        return doc

    def dump_to_blob(self, doc):
        blob = self.dump_doc(doc)
        blob[self.TYPE_KEY] = self._get_name_from_class(type(doc))
        return blob


class Conflict(Exception):
    """
    There was a conflict when trying to perform the operation.
    """


class Missing(Exception):
    """
    Could not find the requested document.

    Note that this is a 404, not a tombstone.
    """


class Deleted(Exception):
    """
    Requested a deleted document.

    Note that this is a document with a tombstone, not a 404.
    """


class TooManyResults(Exception):
    """
    Requested a single document in a find operation and found more than one.

    Note this is based on number of results. Request was successful.
    """


class FindWarning(UserWarning):
    """
    Warnings reported by the CouchDB _find endpoint.
    """


class CouchSession:
    """
    A connection to CouchDB.

    You probably want to override like::

        class MySession(CouchSession):
            loader = MyRegistry

    """

    _client: httpx.AsyncClient
    _root: httpx.URL

    #: Class responsible for de/serializing data.
    loader: type[DocumentLoader]

    def __init__(self, client: httpx.AsyncClient, root: httpx.URL):
        self._client = client
        self._root = root

    @staticmethod
    def _fix_params(params):
        rv = {}
        for key, value in params.items():
            if value is None:
                continue
            elif isinstance(value, str):
                rv[key] = value
            else:
                rv[key] = json.dumps(value)
        return rv

    async def _request(self, method, *urlparts, **kwargs):
        url = self._root.join("/".join(urlparts))
        if "params" in kwargs:
            kwargs["params"] = self._fix_params(kwargs["params"])
        resp = await self._client.request(method, url, **kwargs)
        try:
            resp.raise_for_status()
        except httpx.HTTPStatusError as exc:
            exc.add_note(f"Body: {exc.response.text}")
            match exc.response.status_code:
                case 404:
                    raise Missing(f"Could not find {'/'.join(urlparts)}") from exc
                case 409:
                    raise Conflict(f"Conflict updating {'/'.join(urlparts)}") from exc
                case _:
                    raise
        return resp

    def __getitem__(self, key: str) -> "Database":
        """
        Gets a database.

        (Does not actually check if it exists.)
        """
        return Database(self, key)

    async def get_db(self, dbname: str) -> "Database":
        """
        Gets a database. Checks if it exists.

        See :http:head:`/{db}`
        """
        await self._request("HEAD", dbname)
        return Database(self, dbname)

    async def create_db(
        self,
        dbname: str,
        *,
        shards: int | None = None,
        replicas: int | None = None,
        partitioned: bool | None = None,
    ) -> "Database":
        """
        Create a database

        See :http:post:`/{db}`
        """
        try:
            await self._request(
                "PUT",
                dbname,
                params={
                    "q": shards,
                    "n": replicas,
                    "partitioned": partitioned,
                },
            )
        except httpx.HTTPStatusError as exc:
            match exc.response.status_code:
                case 412:
                    raise Conflict("Database already exists") from exc
                case _:
                    raise
        return Database(self, dbname)

    async def delete_db(self, dbname: str):
        """
        Delete a database.

        See :http:delete:`/{db}`
        """
        await self._request("DELETE", dbname)

    async def iter_dbs(self) -> AsyncIterator[str]:
        """
        List all databases

        See :http:get:`/_all_dbs`
        """
        resp = await self._request("GET", "_all_dbs")
        for dbname in resp.json():
            yield dbname

    # TODO: Database metadata


class Database:
    """
    An individual database.
    """

    def __init__(self, session, name):
        """
        :private:
        """
        self._session = session
        self._name = name

    def _blob2doc(self, blob, db, docid, etag=...):
        if docid is ...:
            docid = blob["_id"]
        if etag is ...:
            etag = f'"{blob["_rev"]}"'
        doc = self._session.loader().load_from_blob(blob)
        doc.__db = db
        doc.__docid = docid
        doc.__etag = etag
        return doc

    def _doc2blob(self, doc):
        blob = self._session.loader().dump_to_blob(doc)
        db = docid = etag = None
        try:
            db = doc.__db
            docid = doc.__docid
            etag = doc.__etag
        except AttributeError:
            pass
        return blob, db, docid, etag

    def _touch_doc(self, doc, *, db=None, etag=None, rev=None, id=None):
        fields = {}
        if etag is not None:
            doc.__etag = etag
        if db is not None:
            doc.__db = db
        if rev is not None:
            fields["_rev"] = rev
        if id is not None:
            doc.__docid = id
            fields["_id"] = id
        if fields:
            self._session.loader().update_doc(doc, **fields)

    async def get(
        self,
        docid: str,
        *,
        attachments: bool = False,
        conflicts: bool = False,
        deleted_conflicts: bool = False,
        latest: bool = False,
        local_seq: bool = False,
        meta: bool = False,
        open_revs: list[str] | Literal["all"] | None = None,
        rev: str | None = None,
        revs: bool = False,
        revs_info: bool = False,
    ):
        """
        Get a document

        See :http:get:`/{db}/{docid}`
        """
        resp = await self._session._request(
            "GET",
            self._name,
            docid,
            params={
                "attachments": attachments,
                "conflicts": conflicts,
                "deleted_conflicts": deleted_conflicts,
                "latest": latest,
                "local_seq": local_seq,
                "meta": meta,
                "open_revs": open_revs,
                "rev": rev,
                "revs": revs,
                "revs_info": revs_info,
            },
            headers={
                "Accept": "application/json",
            },
        )

        blob = resp.json()
        if blob.get("_deleted", False):  # TODO: Flag to override this
            raise Deleted("Document {self._name}/{docid} is marked as deleted")
        if "ETag" in resp.headers:
            etag = resp.headers["ETag"]
        else:
            # Conflicts mode
            etag = f'"{blob["_rev"]}"'

        # For some reason, requested fiels are omitted when empty
        if attachments:
            blob.setdefault("_attachments", {})
        if conflicts:
            blob.setdefault("_conflicts", [])
        if deleted_conflicts:
            blob.setdefault("_deleted_conflicts", [])
        if revs:
            blob.setdefault("_revisions", {})
        if revs_info or open_revs:
            blob.setdefault("_revisions", {})

        doc = self._blob2doc(blob, self._name, docid, etag)
        return doc

    # TODO: Attachments

    async def find_one(
        self, selector: typing.Mapping, use_index: str | list[str] | None = None
    ):
        """
        Get a single document based on ``selector``.

        See :http:post:`/{db}/_find`
        """
        json_body = {"selector": selector, "limit": 2}

        if use_index is not None:
            json_body |= {"use_index": use_index}

        resp = await self._session._request("POST", self._name, "_find", json=json_body)

        blob = resp.json()
        results = blob.get("docs")
        match len(results):
            case 0:
                raise Missing("No results found.")
            case 1:
                return self._blob2doc(results[0], self._name, ...)
            case _:
                raise TooManyResults("More than one result found.")

    async def find(
        self,
        selector: typing.Mapping,
        use_index: str | list[str] | None = None,
        pagesize: int | None = None,
    ) -> AsyncIterator:
        """
        Generate documents based on ``selector``.

        See :http:post:`/{db}/_find`
        """
        json_body: dict[str, Any] = {"selector": selector, "bookmark": None}

        if use_index is not None:
            json_body |= {"use_index": use_index}

        if pagesize is not None:
            json_body |= {"limit": pagesize}

        while True:
            resp = await self._session._request(
                "POST", self._name, "_find", json=json_body
            )
            payload = resp.json()
            if payload.get("warning", None):
                warnings.warn(payload["warning"], FindWarning)

            for doc in payload["docs"]:
                yield self._blob2doc(doc, self._name, ...)

            if not payload["docs"]:
                break

            json_body["bookmark"] = payload["bookmark"]

    async def attempt_put(
        self,
        doc,
        docid: str | None = None,
        *,
        batch: bool = False,
    ):
        """
        Update a document.

        db and docid only need to be given if it's a new document.

        See :http:put:`/{db}/{docid}`
        """
        blob, _db, _docid, etag = self._doc2blob(doc)
        assert _db is None or _db == self._name
        resp = await self._session._request(
            "PUT",
            self._name,
            _docid or docid,
            params={"batch": "ok"} if batch else {},
            headers={"If-Match": etag} if etag else {},
            json=blob,
        )
        payload = resp.json()
        assert payload["ok"]
        self._touch_doc(
            doc,
            db=self._name,
            id=payload["id"],
            etag=resp.headers["ETag"],
            rev=resp.json()["rev"],
        )

    async def attempt_delete(self, doc, *, batch: bool = False):
        """
        Delete a document

        See :http:delete:`/{db}/{docid}`
        """
        _, db, docid, etag = self._doc2blob(doc)
        assert db == self._name
        assert docid
        await self._session._request(
            "DELETE",
            db,
            docid,
            params={"batch": "ok"} if batch else {},
            headers={"If-Match": etag},
        )

    async def attempt_copy(self, src_doc, dst_doc, *, batch: bool = False):
        """
        Copy a document

        .. todo::

           Implement

        See :http:copy:`/{db}/{docid}`
        """
        # FIXME: Figure out signature

    async def mutate(self, docid: str) -> AsyncIterator:
        """
        A document mutation loop::

            async for doc in couch.mutate_doc("spam"):
                doc.foo = "bar"

        Will replay the mutation until it goes through.
        """
        doc = await self.get(docid)
        while True:
            yield doc
            try:
                await self.attempt_put(doc)
            except Conflict:
                doc = await self.get(docid)
            else:
                break

    async def iter_all_docs(
        self, *, include_docs: bool = False
    ) -> AsyncIterator[structs.AllDocs_DocRef]:
        """
        List all documents.

        This excludes design documents, see :meth:`.iter_design_docs`.

        Args:
            include_docs: Pre-load documents

        See :http:get:`/{db}/_all_docs`
        """
        # TODO: Pagination
        resp = await self._session._request(
            "GET",
            self._name,
            "_all_docs",
            params={
                "include_docs": include_docs,
            },
            headers={
                "Accept": "application/json",
            },
        )
        blob = resp.json()
        for ref in blob["rows"]:
            if ref["id"].startswith("_design/"):
                continue
            if "doc" in ref:
                doc = self._blob2doc(ref["doc"], self._name, ref["id"])
            else:
                doc = None
            yield structs.AllDocs_DocRef(
                _db=self, docid=ref["id"], rev=ref["value"]["rev"], _doc=doc
            )

    async def iter_indexes(self) -> AsyncIterator[structs.Index]:
        resp = await self._session._request("GET", self._name, "_index")
        payload = resp.json()
        for idx in payload["indexes"]:
            if idx["ddoc"] is None and idx["name"] == "_all_docs":
                continue
            yield structs.Index.from_dict(idx)

    async def add_index(
        self,
        name: str | None = None,
        *,
        fields: list[str],
        ddoc: str | None = None,
        type: str = "json",
    ):
        await self._session._request(
            "POST",
            self._name,
            "_index",
            json={
                "index": {"fields": list(fields)},
                "name": name,
                "ddoc": ddoc,
                "type": type,
            },
        )


class NoServerFound(Exception):
    """
    None of the configured servers seem to be working.
    """


class SessionPool:
    """
    Responsible for giving out Couch connections.

    You probably want to override like::

        class MyPool(SessionPool):
            session_class = MySession
    """

    _client: httpx.AsyncClient

    #: Class to use for sessions
    session_class: type[CouchSession]

    def __init__(self):
        super().__init__()
        self._client = self.make_client()

    def make_client(self) -> httpx.AsyncClient:
        """
        Produce an httpx client.
        """
        return httpx.AsyncClient(http2=True, follow_redirects=True)

    async def iter_servers(self) -> AsyncIterator[str | httpx.URL]:
        """
        Produce the list of potential servers.

        Override this
        """
        raise NotImplementedError
        for _ in ():
            yield

    async def _check_server(self, url: httpx.URL):
        resp = await self._client.get(url.join("_up"))
        resp.raise_for_status()

    async def session(self) -> CouchSession:
        """
        Get a session
        """
        excs = []
        async for url in self.iter_servers():
            url = httpx.URL(url)
            try:
                await self._check_server(url)
            except Exception as e:
                excs.append(e)
            else:
                return self.session_class(self._client, url)
        else:
            raise NoServerFound() from ExceptionGroup(
                "There were errors checking servers", excs
            )



"""
Provides utilities for testing
"""

import contextlib
import functools
from pathlib import Path
import socket
import sys
import time
import typing

import anyio
import docker
import docker.utils
import httpx


class ContextNotExistError(ValueError):
    """
    The given Docker context does not exist.

    This should only happen if your Docker client is misconfigured.
    """


def _docker_3190_workaround():
    """
    Work around for https://github.com/docker/docker-py/issues/3190
    """
    if docker.utils.config.find_config_file() is None:
        # TODO: Prefer .config_path_from_environment() over .home_dir()
        config_path = (
            Path(docker.utils.config.home_dir())
            / docker.utils.config.DOCKER_CONFIG_FILENAME
        )

        if config_path.parent.exists():
            # If .docker doesn't exist, it doesn't contain contexts
            config_path.touch()


@functools.cache
def _get_docker_client(use: str | None = None) -> docker.DockerClient:
    """
    Get a docker client for the given docker context.

    Unlike docker.from_env(), this considers the user's configured context.
    """
    _docker_3190_workaround()

    context = docker.ContextAPI.get_context(use)
    if context is None:
        raise ContextNotExistError(f"Docker context {use!r} not found")
    return docker.DockerClient(
        base_url=context.endpoints["docker"]["Host"], tls=context.TLSConfig
    )


# https://stackoverflow.com/a/45690594
def _find_free_port():
    with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


@contextlib.contextmanager
def spawn_docker_couchdb() -> typing.Iterator[str]:
    """
    Creates a tempory CouchDB instance using Docker, and automatically cleans it up.

    Returns the URL by which it's accessible.
    """
    client = _get_docker_client()

    couch_container = client.containers.run(
        detach=True,
        image="ghcr.io/teahouse-hosting/quick-and-dirty-couch:latest",
        auto_remove=True,
        environment={
            "COUCHDB_USER": "admin",
            "COUCHDB_PASSWORD": "admin",
        },
        ports={"5984/tcp": None},
    )
    # TODO: Stream container stdout

    port_config = None
    while port_config is None:
        couch_container.reload()
        try:
            # Dig out the connected port
            port_config = couch_container.attrs["NetworkSettings"]["Ports"]["5984/tcp"][
                0
            ]
        except IndexError:
            time.sleep(0.1)

    couch_ip = port_config["HostIp"]
    if couch_ip == "0.0.0.0":
        couch_ip = "127.0.0.1"
    elif couch_ip == "::":
        couch_ip = "::1"
    couch_port = port_config["HostPort"]

    try:
        yield f"http://admin:admin@{couch_ip}:{couch_port}/"
    finally:
        couch_container.stop()


async def wait_for_readiness(couch_url: str, *, timeout: float = 60) -> None:
    """
    Waits for a CouchDB instance to actually initialize and be ready to accept
    requests.

    Raises TimeoutError on failure.
    """
    timeout_increment = 0.1
    url = httpx.URL(couch_url).join("_up")
    client = httpx.AsyncClient()
    for _ in range(int(timeout / timeout_increment)):
        try:
            resp = await client.get(url)
        except httpx.RequestError:
            await anyio.sleep(timeout_increment)
        else:
            if resp.is_success:
                return
            else:
                await anyio.sleep(timeout_increment)
    else:
        raise TimeoutError(f"Timeout waiting for CouchDB to initialize ({couch_url})")


async def _call_cli(couch_url, *argv):
    oldargv = sys.argv
    try:
        sys.argv = ["chaise", "--verbose", "--server", couch_url, *argv]
        import chaise.cli

        await chaise.cli.main()
    finally:
        sys.argv = oldargv


@contextlib.asynccontextmanager
async def run_cli_apply(couch_url: str, dbs_module: str):
    """
    Context manager that runs the apply CLI command, and then cleans up
    databases afterwards.
    """
    await _call_cli(couch_url, "apply", dbs_module)
    yield
    # TODO: Look up what databases were actually defined
    from chaise.cli.client import ConstantPool

    session = await ConstantPool(couch_url).session()
    dbs = {db async for db in session.iter_dbs() if not db.startswith("_")}
    for db in dbs:
        await session.delete_db(db)



"""
Extra bits to make your job easier.
"""

import random
from socket import AddressFamily, SocketKind


import anyio
import httpx

from . import SessionPool


class ConstantPoolMixin(SessionPool):
    """
    A trivial pool mixin that uses the passed in URL(s).
    """

    urls: list[str]

    def __init__(self, urls: str | list[str]):
        if isinstance(urls, str):
            self.urls = [urls]
        else:
            self.urls = list(urls)
        super().__init__()

    async def iter_servers(self):
        u = self.urls[:]
        random.shuffle(u)
        for url in u:
            yield url


async def _get_ips(hostname: str | bytes, port: int | None = None):
    """
    Look up the IPs for a given hostname
    """
    for gai in await anyio.getaddrinfo(hostname, port):
        match gai:
            case (AddressFamily.AF_INET, SocketKind.SOCK_STREAM, _, _, (addr, _)):
                yield addr
            case (
                AddressFamily.AF_INET6,
                SocketKind.SOCK_STREAM,
                _,
                _,
                (addr, _),
            ):
                yield addr
            case (
                AddressFamily.AF_INET6,
                SocketKind.SOCK_STREAM,
                _,
                _,
                (addr, _, _, scope),
            ):
                # IPv6 flow is without purpose and cannot be represented in addresses
                if scope:
                    yield f"{addr}%{scope}"
                else:
                    yield addr
            case _:
                continue


class DnsPoolMixin(SessionPool):
    """
    Uses DNS round robin to find pool members.

    Assumes that CouchDB is not behind a vhost or uses TLS.
    """

    url: httpx.URL

    def __init__(self, url: str):
        self.url = httpx.URL(url)
        super().__init__()

    async def iter_servers(self):
        # The domain stack already does our shuffling
        async for ip in _get_ips(self.url.raw_host, self.url.port):
            if ":" in ip:
                ip = f"[{ip}]"
            yield str(self.url.copy_with(host=ip))



"""
Integration for attrs/cattrs
"""

import typing
from typing import AbstractSet
import warnings

import attrs
import cattrs.converters
import cattrs.gen

try:
    # ujson is preferred, since muffin&c can also use it
    from cattrs.preconf.ujson import configure_converter
except ImportError:
    from cattrs.preconf.json import configure_converter
# Omitting orjson, even though it's a preconf, because I'm not confident it's a
# drop-in equivalent to (u)json

from . import DocumentRegistry


# All implementations exhibit the conversions:
# * bytes are wrapped in base85
# * dates & datetimes are ISO 8601
#: The converter used when talking to CouchDB.
converter = cattrs.converters.Converter(
    unstruct_collection_overrides={
        AbstractSet: list,
    }
)
configure_converter(converter)


class classprop:
    """
    Like @property, but for class attributes

    :meta private:
    """

    def __init__(self, factory: typing.Callable):
        self._factory = factory
        self.__doc__ = factory.__doc__

    def __get__(self, instance, owner):
        return self._factory(owner)

    def __set__(self, instance, value):
        raise AttributeError("Cannot set a classprop")

    def __set_name__(self, owner, name):
        self.__objclass__ = owner


class DocMeta(type):
    """
    Defines an attrs class as a subclass instead of a decorator

    :meta private:
    """

    def __new__(cls, name, bases, dict, **kwds):
        sub = super().__new__(cls, name, bases, dict)
        dbid = kwds.pop("dbid", None)
        assert not kwds.get("slots", None)
        if "__attrs_attrs__" not in dict:  # prevents recursion
            sub = attrs.define(**kwds)(sub)

            # Register the class with chaise
            if sub.__module__ != globals()["__name__"] and dbid is None:
                warnings.warn(f"No dbid given for {sub!r}")
            if dbid is not None and sub._Document__parent is not None:
                sub._Document__parent.document(dbid)(sub)

            # Register the class with cattrs
            converter.register_unstructure_hook(
                sub,
                cattrs.gen.make_dict_unstructure_fn(
                    sub,
                    converter,
                    _id=cattrs.gen.override(omit=True),
                    _rev=cattrs.gen.override(omit=True),
                    _deleted=cattrs.gen.override(omit_if_default=True),
                    _attachments=cattrs.gen.override(omit=True),
                    _conflicts=cattrs.gen.override(omit=True),
                    _deleted_conflicts=cattrs.gen.override(omit=True),
                    _local_seq=cattrs.gen.override(omit=True),
                    _revs_info=cattrs.gen.override(omit=True),
                    _revisions=cattrs.gen.override(omit=True),
                ),
            )
        return sub


class Document(metaclass=DocMeta, slots=False, frozen=False):
    """
    The parent class for all documents that get saved to Couch.

    Do not inherit from directly; use :class:`AttrsRegistry.Document` instead.
    """

    __parent: typing.ClassVar[type | None] = None

    #: Document ID
    #:
    #: :meta public:
    _id: str | None = attrs.field(default=None, kw_only=True, alias="_id")

    #: Document revision
    #:
    #: :meta public:
    _rev: str | None = attrs.field(default=None, kw_only=True, alias="_rev")

    #: Has the document been deleted? (ie, is this a tombstone?)
    #:
    #: :meta public:
    _deleted: bool = attrs.field(default=False, kw_only=True, alias="_deleted")

    #: Attachment information, if requested
    #:
    #: :meta public:
    _attachments: dict | None = attrs.field(
        default=None, kw_only=True, alias="_attachments"
    )

    #: List of alternate document versions, if requested
    #:
    #: :meta public:
    _conflicts: list | None = attrs.field(
        default=None, kw_only=True, alias="_conflicts"
    )

    #: List of deleted conflicts, if requested
    #:
    #: :meta public:
    _deleted_conflicts: list | None = attrs.field(
        default=None, kw_only=True, alias="_deleted_conflicts"
    )

    #:
    #: :meta public:
    _local_seq: str | None = attrs.field(default=None, kw_only=True, alias="_local_seq")

    #:
    #: :meta public:
    _revs_info: list | None = attrs.field(
        default=None, kw_only=True, alias="_revs_info"
    )

    #:
    #: :meta public:
    _revisions: dict | None = attrs.field(
        default=None, kw_only=True, alias="_revisions"
    )


class AttrsRegistry(DocumentRegistry):
    @classprop
    def Document(cls) -> type[Document]:
        """
        Document superclass.

        Handles:

        * Making an attrs
        * Registering with chaise
        * Serialization concerns

        Me sure to include the ``dbid`` keyword argument::

           class MyDoc(AttrsRegistry.Document, dbid="mydoc"): ...
        """

        # This is some shenanigans because names
        class Document(globals()["Document"]):  # type: ignore
            __doc__ = vars(AttrsRegistry)["Document"].__doc__
            __parent = cls

        return Document

    def load_doc(self, cls: type, blob: dict):
        """
        :meta private:
        """
        return converter.structure(blob, cls)

    def dump_doc(self, doc) -> dict:
        """
        :meta private:
        """
        return converter.unstructure(doc)

    def update_doc(self, doc, **fields):
        assert all(k.startswith("_") for k in fields)
        for k, v in fields.items():
            setattr(doc, k, v)
